/* $Id: uaccess.h,v 1.3 1999/10/12 14:46:20 gniibe Exp $
 *
 * User space memory access functions
 *
 * Copyright (C) 1999  Niibe Yutaka
 *
 *  Based on:
 *     MIPS implementation version 1.15 by
 *              Copyright (C) 1996, 1997, 1998 by Ralf Baechle
 *     and i386 version.
 */
#ifndef __ASM_SH_UACCESS_H
#define __ASM_SH_UACCESS_H

#include <linux/errno.h>
#include <linux/sched.h>

#ifdef CONFIG_ZAURUS
/* uso uso */
extern long dbg_mem_get1(long,long);
extern long dbg_mem_get4(long,long);
extern long dbg_mem_put1(long,long,long);
extern long dbg_mem_put4(long,long,long);
//static unsigned long target_procname(struct task_struct*);
static __inline__ unsigned long target_procname(struct task_struct *task)
{
	return task->thread.procname;
}
#if 0
#define CURRENT_PROC_DEBUG0(m)	printk("%s: 0\n", (m))
#define CURRENT_PROC_DEBUG(m, c)	printk("%s: cur = %lx\n", (m), (c))
#define CURRENT_PROC_DEBUG2(m, c)	printk("%s: ret = %lx\n", (m), (unsigned long)(c))
#else
#define CURRENT_PROC_DEBUG0(m)	
#define CURRENT_PROC_DEBUG(m, c)
#define CURRENT_PROC_DEBUG2(m, c)
#endif
#endif CONFIG_ZAURUS

#define VERIFY_READ    0
#define VERIFY_WRITE   1

/*
 * The fs value determines whether argument validity checking should be
 * performed or not.  If get_fs() == USER_DS, checking is performed, with
 * get_fs() == KERNEL_DS, checking is bypassed.
 *
 * For historical reasons (Data Segment Register?), these macros are misnamed.
 */

#define MAKE_MM_SEG(s)	((mm_segment_t) { (s) })

#define KERNEL_DS	MAKE_MM_SEG(0xFFFFFFFF)
#define USER_DS		MAKE_MM_SEG(0x80000000)

#define get_ds()	(KERNEL_DS)
#define get_fs()        (current->addr_limit)
#define set_fs(x)       (current->addr_limit=(x))

#define segment_eq(a,b)	((a).seg == (b).seg)

#define __addr_ok(addr) ((unsigned long)(addr) < (current->addr_limit.seg))

/*
 * Uhhuh, this needs 33-bit arithmetic. We have a carry..
 *
 * sum := addr + size;  carry? --> flag = true;
 * if (sum >= addr_limit) flag = true;
 */
#if 1227
#define __range_ok(addr,size) (addr == 0 || (addr + size) == 0)
#else
#define __range_ok(addr,size) ({ \
	unsigned long flag,sum; \
	__asm__("clrt; addc %3,%1; movt %0; cmp/hi %4,%1; rotcl %0" \
		:"=&r" (flag), "=r" (sum) \
		:"1" (addr), "r" ((int)(size)), "r" (current->addr_limit.seg)); \
	flag; })
#endif

#define access_ok(type,addr,size) (__range_ok(addr,size) == 0)
#define __access_ok(addr,size) (__range_ok(addr,size) == 0)

extern inline int verify_area(int type, const void * addr, unsigned long size)
{
	return access_ok(type,addr,size) ? 0 : -EFAULT;
}

/*
 * Uh, these should become the main single-value transfer routines ...
 * They automatically use the right size if we just have the right
 * pointer type ...
 *
 * As MIPS uses the same address space for kernel and user data, we
 * can just do these as direct assignments.
 *
 * Careful to not
 * (a) re-use the arguments for side effects (sizeof is ok)
 * (b) require any knowledge of processes at this stage
 */
#define put_user(x,ptr)	__put_user_check((x),(ptr),sizeof(*(ptr)))
#define get_user(x,ptr) __get_user_check((x),(ptr),sizeof(*(ptr)))

/*
 * The "__xxx" versions do not do address space checking, useful when
 * doing multiple accesses to the same area (the user has to do the
 * checks by hand with "access_ok()")
 */
#define __put_user(x,ptr) __put_user_nocheck((x),(ptr),sizeof(*(ptr)))
#define __get_user(x,ptr) __get_user_nocheck((x),(ptr),sizeof(*(ptr)))

/*
 * The "xxx_ret" versions return constant specified in third argument, if
 * something bad happens. These macros can be optimized for the
 * case of just returning from the function xxx_ret is used.
 */

#define put_user_ret(x,ptr,ret) ({ \
if (put_user(x,ptr)) return ret; })

#define get_user_ret(x,ptr,ret) ({ \
if (get_user(x,ptr)) return ret; })

#define __put_user_ret(x,ptr,ret) ({ \
if (__put_user(x,ptr)) return ret; })

#define __get_user_ret(x,ptr,ret) ({ \
if (__get_user(x,ptr)) return ret; })

struct __large_struct { unsigned long buf[100]; };
#define __m(x) (*(struct __large_struct *)(x))

#ifdef CONFIG_ZAURUS
/*** ASSUMPTION: must be little endian!!! ***/

#define GET_USER_ASSERT(x)					\
do {								\
	if (!(x))						\
		printk ("get_user assersion failed in %s:%d\n",	\
			__FILE__, __LINE__);			\
} while (0)

#define __get_user_xtal(a) ({						\
	if ((__gu_tmp = dbg_mem_get1(__cur, __gu_addr+(a))) == -1)	\
		__gu_err = -EFAULT;					\
	__gu_tmp;							\
})

#define __get_user_nocheck(x,ptr,size) ({				    \
	long __gu_err = 0;						    \
	unsigned long __gu_val = 0;					    \
	unsigned long __gu_addr = (unsigned long)(ptr);			    \
	if (segment_eq(get_fs(), get_ds())){				    \
		switch(size){						    \
		default:						    \
			__gu_err = -EFAULT;				    \
			break;						    \
		case 4:							    \
			__gu_val |= *((unsigned char*)__gu_addr + 2) << 16; \
			__gu_val |= *((unsigned char*)__gu_addr + 3) << 24; \
		case 2:							    \
			__gu_val |= *((unsigned char*)__gu_addr + 1) << 8;  \
		case 1:							    \
			__gu_val |= *(unsigned char*)__gu_addr;		    \
			break;						    \
		}							    \
	} else {							    \
		long __cur = target_procname(current);			    \
		unsigned long __gu_tmp;					    \
		GET_USER_ASSERT(__cur != 0);				    \
		switch (size) {						    \
		default:						    \
			__gu_err = -EFAULT;				    \
			break;						    \
		case 4:							    \
			__gu_val |= __get_user_xtal(2) << 16;		    \
			__gu_val |= __get_user_xtal(3) << 24;		    \
		case 2:							    \
			__gu_val |= __get_user_xtal(1) << 8;		    \
		case 1:							    \
			__gu_val |= __get_user_xtal(0);			    \
			break;						    \
		}							    \
	}								    \
	if (!__gu_err)							    \
		x = __gu_val;						    \
	__gu_err;							    \
})

#define __get_user_check(x,ptr,size) ({				\
	int retval;						\
	if (__access_ok((unsigned long)(ptr),(size))){		\
		retval = __get_user_nocheck((x),(ptr),(size));	\
	} else {						\
		retval = -EFAULT;				\
	}							\
	retval;							\
})


#define __put_user_nocheck(x,ptr,size) ({				    \
	long __pu_err = 0;						    \
	unsigned long __pu_val = (unsigned long)(x);			    \
	unsigned long __pu_addr = (unsigned long)(ptr);			    \
	if (segment_eq(get_fs(), get_ds())){				    \
		switch (size){						    \
		default:						    \
			__pu_err = -EFAULT;				    \
			break;						    \
		case 4:							    \
			*((unsigned char*)__pu_addr+2)=(__pu_val>>16)&0xff; \
			*((unsigned char*)__pu_addr+3)=(__pu_val>>24)&0xff; \
		case 2:							    \
			*((unsigned char*)__pu_addr+1)=(__pu_val>> 8)&0xff; \
		case 1:							    \
			*((unsigned char*)__pu_addr  )= __pu_val     &0xff; \
			break;						    \
		}							    \
	} else {							    \
		long __cur = target_procname(current);			    \
		GET_USER_ASSERT(__cur != 0);				    \
		switch (size){						    \
		default:						    \
			__pu_err = -EFAULT;				    \
			break;						    \
		case 4:							    \
			if (dbg_mem_put4(__cur, __pu_addr, __pu_val) == -1) \
				__pu_err = -EFAULT;			    \
			break;						    \
		case 2:							    \
			if (dbg_mem_put1(__cur, __pu_addr+1, (__pu_val>>8)&0xff) == -1){    \
				__pu_err = -EFAULT;			    \
				break;					    \
			}						    \
		case 1:							    \
			if (dbg_mem_put1(__cur, __pu_addr, __pu_val&0xff) == -1)   \
				__pu_err = -EFAULT;			    \
			break;						    \
		}							    \
	}								    \
	__pu_err;							    \
})

#define __put_user_check(x,ptr,size) ({				\
	int retval;						\
	if (__access_ok((unsigned long)(ptr), (size))){		\
		retval = __put_user_nocheck((x),(ptr),(size));	\
	} else {						\
		retval = -EFAULT;				\
	}							\
	retval;							\
})

#else CONFIG_ZAURUS

#define __get_user_nocheck(x,ptr,size) ({ \
long __gu_err; \
__typeof(*(ptr)) __gu_val; \
long __gu_addr; \
__asm__("":"=r" (__gu_val)); \
__gu_addr = (long) (ptr); \
__asm__("":"=r" (__gu_err)); \
switch (size) { \
case 1: __get_user_asm("b"); break; \
case 2: __get_user_asm("w"); break; \
case 4: __get_user_asm("l"); break; \
default: __get_user_unknown(); break; \
} x = (__typeof__(*(ptr))) __gu_val; __gu_err; })

#define __get_user_check(x,ptr,size) ({ \
long __gu_err; \
__typeof__(*(ptr)) __gu_val; \
long __gu_addr; \
__asm__("":"=r" (__gu_val)); \
__gu_addr = (long) (ptr); \
__asm__("":"=r" (__gu_err)); \
if (__access_ok(__gu_addr,size)) { \
switch (size) { \
case 1: __get_user_asm("b"); break; \
case 2: __get_user_asm("w"); break; \
case 4: __get_user_asm("l"); break; \
default: __get_user_unknown(); break; \
} } x = (__typeof__(*(ptr))) __gu_val; __gu_err; })

#define __get_user_asm(insn) \
({ \
__asm__ __volatile__( \
	"1:\n\t" \
	"mov." insn "	%2,%1\n\t" \
	"mov	#0,%0\n" \
	"2:\n\t" \
	".section	.fixup,\"ax\"\n" \
	"3:\n\t" \
	"mov	#0,%1\n\t" \
	"mov.l	4f,%0\n\t" \
	"jmp	@%0\n\t" \
	" mov	%3,%0\n" \
	"4:	.long	2b\n\t" \
	".previous\n\t" \
	".section	__ex_table,\"a\"\n\t" \
	".long	1b, 3b\n\t" \
	".previous" \
	:"=&r" (__gu_err), "=&r" (__gu_val) \
	:"m" (__m(__gu_addr)), "i" (-EFAULT)); })

extern void __get_user_unknown(void);

#define __put_user_nocheck(x,ptr,size) ({ \
long __pu_err; \
__typeof__(*(ptr)) __pu_val; \
long __pu_addr; \
__pu_val = (x); \
__pu_addr = (long) (ptr); \
__asm__("":"=r" (__pu_err)); \
switch (size) { \
case 1: __put_user_asm("b"); break; \
case 2: __put_user_asm("w"); break; \
case 4: __put_user_asm("l"); break; \
default: __put_user_unknown(); break; \
} __pu_err; })

#define __put_user_check(x,ptr,size) ({ \
long __pu_err; \
__typeof__(*(ptr)) __pu_val; \
long __pu_addr; \
__pu_val = (x); \
__pu_addr = (long) (ptr); \
__asm__("":"=r" (__pu_err)); \
if (__access_ok(__pu_addr,size)) { \
switch (size) { \
case 1: __put_user_asm("b"); break; \
case 2: __put_user_asm("w"); break; \
case 4: __put_user_asm("l"); break; \
default: __put_user_unknown(); break; \
} } __pu_err; })

#define __put_user_asm(insn) \
({ \
__asm__ __volatile__( \
	"1:\n\t" \
	"mov." insn "	%1,%2\n\t" \
	"mov	#0,%0\n" \
	"2:\n\t" \
	".section	.fixup,\"ax\"\n" \
	"3:\n\t" \
	"nop\n\t" \
	"mov.l	4f,%0\n\t" \
	"jmp	@%0\n\t" \
	"mov	%3,%0\n" \
	"4:	.long	2b\n\t" \
	".previous\n\t" \
	".section	__ex_table,\"a\"\n\t" \
	".long	1b, 3b\n\t" \
	".previous" \
	:"=&r" (__pu_err) \
	:"r" (__pu_val), "m" (__m(__pu_addr)), "i" (-EFAULT)); })

extern void __put_user_unknown(void);

#endif CONFIG_ZAURUS


/* Generic arbitrary sized copy.  */
/* XXX: should be such that: 4byte and the rest. */
#ifndef CONFIG_ZAURUS
extern __inline__ __kernel_size_t
__copy_user(void *__to, const void *__from, __kernel_size_t __n)
{
	unsigned long __dummy, _f, _t;
	__kernel_size_t res;

	__asm__ __volatile__(
		"9:\n\t"
		"mov.b	@%2+,%1\n\t"
		"dt	%0\n"
		"1:\n\t"
		"mov.b	%1,@%3\n\t"
		"bf/s	9b\n\t"
		" add	#1,%3\n"
		"2:"
		".section .fixup,\"ax\"\n"
		"3:\n\t"
		"mov.l	5f,%1\n\t"
		"jmp	@%1\n\t"
		" mov	%7,%0\n\t"
		".balign 4\n"
		"5:	.long 2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.balign 4\n"
		"	.long 9b,3b\n"
		"	.long 1b,2b\n"
		".previous"
		: "=&r" (res), "=&z" (__dummy), "=&r" (_f), "=&r" (_t)
		: "2" (__from), "3" (__to), "0" (__n), "i" (-EFAULT)
		: "memory");

	return res;
}
#endif !CONFIG_ZAURUS

#ifdef CONFIG_ZAURUS

#define copy_to_user(to,from,n) ({ \
void *__copy_to = (void *) (to); \
__kernel_size_t __copy_size = (__kernel_size_t) (n); \
__kernel_size_t __copy_res; \
if(__copy_size && __access_ok((unsigned long)__copy_to, __copy_size)) { \
__copy_res = __copy_to_user(__copy_to, (void *) (from), __copy_size); \
} else __copy_res = __copy_size; \
__copy_res; })

#else CONFIG_ZAURUS

#define copy_to_user(to,from,n) ({ \
void *__copy_to = (void *) (to); \
__kernel_size_t __copy_size = (__kernel_size_t) (n); \
__kernel_size_t __copy_res; \
if(__copy_size && __access_ok((unsigned long)__copy_to, __copy_size)) { \
__copy_res = __copy_user(__copy_to, (void *) (from), __copy_size); \
} else __copy_res = __copy_size; \
__copy_res; })

#endif CONFIG_ZAURUS

#define copy_to_user_ret(to,from,n,retval) ({ \
if (copy_to_user(to,from,n)) \
	return retval; \
})

#ifdef CONFIG_ZAURUS

/*** very very slow implementation ***/
extern __inline__ __kernel_size_t
__copy_to_user(void *__to, const void *__from, __kernel_size_t __n)
{
	unsigned long _t;
	char *_f;
	long __ret;
	char __val;
	unsigned long __cur;
	CURRENT_PROC_DEBUG0("__copy_to_user");
	__cur = target_procname(current);
// if(__cur == 0){
if (segment_eq(get_fs(), get_ds())){  // kernel to kernel copy
	CURRENT_PROC_DEBUG("__copy_to_user (ker)", __cur);

	_t = (unsigned long)__to;
	_f = (char *)__from;

	while (__n) {
		*(unsigned char*)_t = *_f++; _t++;
		__n--;
	}
	CURRENT_PROC_DEBUG2("__copy_to_user (ker)", __n);
	return __n;
}                            // kernel to user copy
	CURRENT_PROC_DEBUG("__copy_to_user", __cur);

	_t = (unsigned long)__to;
	_f = (char *)__from;

	while (__n) {
		__val = *_f++;	/* sorry, not check exception */
		__ret = dbg_mem_put1(__cur, _t++, __val);
		if (__ret == -1) break;
		__n--;
	}
	CURRENT_PROC_DEBUG2("__copy_to_user", __n);
	return __n;
}

#else CONFIG_ZAURUS

#define __copy_to_user(to,from,n)		\
	__copy_user((void *)(to),		\
		    (void *)(from), n)

#endif CONFIG_ZAURUS

#define __copy_to_user_ret(to,from,n,retval) ({ \
if (__copy_to_user(to,from,n)) \
	return retval; \
})

#ifdef CONFIG_ZAURUS

#define copy_from_user(to,from,n) ({ \
void *__copy_to = (void *) (to); \
void *__copy_from = (void *) (from); \
__kernel_size_t __copy_size = (__kernel_size_t) (n); \
__kernel_size_t __copy_res; \
if(__copy_size && __access_ok((unsigned long)__copy_from, __copy_size)) { \
__copy_res = __copy_from_user(__copy_to, __copy_from, __copy_size); \
} else __copy_res = __copy_size; \
__copy_res; })

#else CONFIG_ZAURUS

#define copy_from_user(to,from,n) ({ \
void *__copy_to = (void *) (to); \
void *__copy_from = (void *) (from); \
__kernel_size_t __copy_size = (__kernel_size_t) (n); \
__kernel_size_t __copy_res; \
if(__copy_size && __access_ok((unsigned long)__copy_from, __copy_size)) { \
__copy_res = __copy_user(__copy_to, __copy_from, __copy_size); \
} else __copy_res = __copy_size; \
__copy_res; })

#endif CONFIG_ZAURUS

#define copy_from_user_ret(to,from,n,retval) ({ \
if (copy_from_user(to,from,n)) \
	return retval; \
})

#ifdef CONFIG_ZAURUS

/*** very very slow implementation ***/
extern __inline__ __kernel_size_t
__copy_from_user(void *__to, const void *__from, __kernel_size_t __n)
{
	char *_t;
	unsigned long _f;
	long __ret;
	long __cur;
	CURRENT_PROC_DEBUG0("__copy_from_user");
	__cur = target_procname(current);
//if(__cur == 0){
if (segment_eq(get_fs(), get_ds())){  // kernel to kernel copy
	CURRENT_PROC_DEBUG("__copy_from_user (ker)", __cur);

	_t = (char *)__to;
	_f = (unsigned long)__from;

	while (__n) {
		*_t++ = *((unsigned char *)_f); _f++;
		__n--;
	}
	CURRENT_PROC_DEBUG2("__copy_from_user (ker)", __n);
	return __n;
}                       // user to kernel copy
	CURRENT_PROC_DEBUG("__copy_from_user", __cur);

	_t = (char *)__to;
	_f = (unsigned long)__from;

	while (__n) {
		__ret = dbg_mem_get1(__cur, _f++);
		if (__ret == -1) return -EFAULT;
		*_t++ = __ret;	/* sorry, not check exception */
		__n--;
	}
	CURRENT_PROC_DEBUG2("__copy_from_user", __n);
	return __n;
}

#else CONFIG_ZAURUS

#define __copy_from_user(to,from,n)		\
	__copy_user((void *)(to),		\
		    (void *)(from), n)

#endif CONFIG_ZAURUS

#define __copy_from_user_ret(to,from,n,retval) ({ \
if (__copy_from_user(to,from,n)) \
	return retval; \
})

/* XXX: Not sure it works well..
   should be such that: 4byte clear and the rest. */
#ifdef CONFIG_ZAURUS

/*** very very slow implementation ***/
extern __inline__ __kernel_size_t
__clear_user(void *addr, __kernel_size_t size)
{
	unsigned long __a, __s;
	long __ret;
	long __cur;
	CURRENT_PROC_DEBUG0("__clear_user");
	printk ("addr = %p, size = %d\n", addr, size);
	__cur = target_procname(current);
// if(__cur == 0){
if (segment_eq(get_fs(), get_ds())){ // handle kernel memory
	CURRENT_PROC_DEBUG("__clear_user (ker)", __cur);

	__a = (unsigned long)addr;
	__s = (unsigned long)size;
	while (__s) {
		*(unsigned char*)__a = 0; __a++;
		__s--;
	}
	CURRENT_PROC_DEBUG2("__clear_user (ker)", size - __s);
	return size - __s;
}                          // handle user memory
	CURRENT_PROC_DEBUG("__clear_user", __cur);

	__a = (unsigned long)addr;
	__s = (unsigned long)size;

	while (__s) {
		__ret = dbg_mem_put1(__cur, __a++, 0);
		if (__ret == -1) {
			CURRENT_PROC_DEBUG2("__clear_user", -EFAULT);
			return -EFAULT;
		}
		__s--;
	}
	CURRENT_PROC_DEBUG2("__clear_user", size - __s);
	return size - __s;
}

#else CONFIG_ZAURUS

extern __inline__ __kernel_size_t
__clear_user(void *addr, __kernel_size_t size)
{
	__kernel_size_t res;
	unsigned long __a, __s;

	__asm__ __volatile__(
		"9:\n\t"
		"dt	%2\n"
		"1:\n\t"
		"mov.b	%5,@%1\n\t"
		"bf/s	9b\n\t"
		" add	#1,%1\n\t"
		"sub	%2,%0\n"
		"2:\n"
		".section .fixup,\"ax\"\n"
		"3:\n\t"
		"mov.l	4f,%0\n\t"
		"jmp	@%0\n\t"
		" mov	%7,%0\n"
		".balign 4\n"
		"4:	.long 2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.balign 4\n"
		"	.long 1b,3b\n"
		".previous"
		: "=&r" (res), "=&r" (__a), "=&r" (__s)
		: "1" (addr), "2" (size), "r" (0), "0" (size), "i" (-EFAULT));

	return res;
}

#endif CONFIG_ZAURUS

#define clear_user(addr,n) ({ \
void * __cl_addr = (addr); \
unsigned long __cl_size = (n); \
if (__cl_size && __access_ok(((unsigned long)(__cl_addr)), __cl_size)) \
__cl_size = __clear_user(__cl_addr, __cl_size); \
__cl_size; })

#ifdef CONFIG_ZAURUS

/*** very very slow implementation ***/
extern __inline__ int
__strncpy_from_user(unsigned long __dest, unsigned long __src, int __count)
{
	__kernel_size_t res;
	unsigned long _s;
	char *_d;
	long __ret;
	long __cur;
	CURRENT_PROC_DEBUG0("__strncpy_from_user");
	__cur = target_procname(current);

//if(__cur == 0){
if (segment_eq(get_fs(), get_ds())){  // kernel to kernel copy
	CURRENT_PROC_DEBUG("__strncpy_from_user (ker)", __cur);
	res = (__kernel_size_t)__count;
	_d = (char *)__dest;
	_s = (unsigned long)__src;
	while (res) {
		*_d++ = __ret = *((unsigned  char *)_s); _s++;
		if (__ret == 0) {
			CURRENT_PROC_DEBUG2("__strncpy_from_user (ker)", res);
			return res;
		}
		res--;
	}
	/* res = res - __count; *//* asm code means...??? */
	res = __count - res; /* should be this code, I think... */
	CURRENT_PROC_DEBUG2("__strncpy_from_user", res);
	return res;
}                          // user to kernel copy
	CURRENT_PROC_DEBUG("__strncpy_from_user", __cur);
	res = (__kernel_size_t)__count;
	_d = (char *)__dest;
	_s = (unsigned long)__src;
//printk("_d=%x _s=%x __count=%d\n",_d,_s,__count);

	/*printk("strncpy_from_user cur=%x\n", __cur);*/
	while (res) {
		/*__ret = *((char *)_s); _s++;*//*debug*/
//printk("dbg_mem_get1(%x,%x)\n",__cur,*((char *)_s));
//printk("dbg_mem_get1(%x,%x)\n",__cur,_s);
		__ret = dbg_mem_get1(__cur, _s++);
//printk("dbg_mem_get1()=%d\n",__ret);
		if (__ret == -1) {
			CURRENT_PROC_DEBUG2("__strncpy_from_user", -EFAULT);
			return -EFAULT;
		}
		*_d++ = __ret;	/* sorry, not check exception */
		if (__ret == 0) {
			CURRENT_PROC_DEBUG2("__strncpy_from_user", res);
			return res;
		}
		res--;
	}
	/* res = res - __count; *//* asm code means...??? */
	res = __count - res; /* should be this code, I think... */
	CURRENT_PROC_DEBUG2("__strncpy_from_user", res);
	return res;
}

#else CONFIG_ZAURUS

extern __inline__ int
__strncpy_from_user(unsigned long __dest, unsigned long __src, int __count)
{
	__kernel_size_t res;
	unsigned long __dummy, _d, _s;

	__asm__ __volatile__(
		"9:\n"
		"mov.b	@%2+,%1\n\t"
		"cmp/eq	#0,%1\n\t"
		"bt/s	2f\n"
		"1:\n"
		"mov.b	%1,@%3\n\t"
		"dt	%0\n\t"
		"bf/s	9b\n\t"
		" add	#1,%3\n\t"
		"sub	%6,%0\n"
		"2:\n"
		".section .fixup,\"ax\"\n"
		"3:\n\t"
		"mov.l	4f,%1\n\t"
		"jmp	@%1\n\t"
		" mov	%8,%0\n\t"
		".balign 4\n"
		"4:	.long 2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.balign 4\n"
		"	.long 9b,3b\n"
		"	.long 1b,2b\n"
		".previous"
		: "=&r" (res), "=&z" (__dummy), "=&r" (_s), "=&r" (_d)
		: "2" (__src), "3" (__dest), "r" (__count), "0" (__count),
		  "i" (-EFAULT)
		: "memory");

	return res;
}

#endif CONFIG_ZAURUS

/*
 * self_sleep(2) in strncpy_from_user:
 *   If __strncpy_from_user accesses *CODE AREA* of user process,
 *   it might be happen that kernel code area is swapped out.
 *   To swap it in, we call self_sleep.
 * TODO:
 *   Check are we accessing user code area and then call self_sleep.
 */
#define strncpy_from_user(dest,src,count) ({ \
unsigned long __sfu_src = (unsigned long) (src); \
int __sfu_count = (int) (count); \
long __sfu_res = -EFAULT; \
if(__access_ok(__sfu_src, __sfu_count)) { \
__sfu_res = __strncpy_from_user((unsigned long) (dest), __sfu_src, __sfu_count); \
self_sleep(2); \
} __sfu_res; })

#define strlen_user(str) strnlen_user(str, ~0UL >> 1)

/*
 * Return the size of a string (including the ending 0!)
 */
#ifdef CONFIG_ZAURUS

/*** very very slow implementation ***/
extern __inline__ long __strnlen_user(const char *__s, long __n)
{
	long __p = (long)__s;
	long __ret;
	long __cur;
	CURRENT_PROC_DEBUG0("__strnlen_user");
	__cur = target_procname(current);

//if(__cur == 0){
if (segment_eq(get_fs(), get_ds())) {  // handle kernel memory
	CURRENT_PROC_DEBUG("__strnlen_user (ker)", __cur);

	while (__n) {
		__ret = *((unsigned char *)__p); __p++;
		if (__ret == 0) break;
		__n--;
	}
	CURRENT_PROC_DEBUG2("__strnlen_user (ker)", __p - (long)__s);
	return __p - (long)__s;
}                             // handle user memory
	CURRENT_PROC_DEBUG("__strnlen_user", __cur);

	while (__n) {
		__ret = dbg_mem_get1(__cur, __p++);
		if (__ret == -1) return -EFAULT;
		if (__ret == 0) break;
		__n--;
	}
	CURRENT_PROC_DEBUG2("__strnlen_user", __p - (long)__s);
	return __p - (long)__s;
}

#else CONFIG_ZAURUS

extern __inline__ long __strnlen_user(const char *__s, long __n)
{
	unsigned long res;
	unsigned long __dummy;

	__asm__ __volatile__(
		"mov	#-1,%1\n"
		"9:\n\t"
		"cmp/eq	%4,%0\n\t"
		"bt	5f\n\t"
		"cmp/eq	#0,%1\n\t"
		"bf/s	9b\n\t"
		"1:\t"
		" mov.b	@%0+,%1\n\t"
		"5:\t"
		"sub	%3,%0\n"
		"2:\n"
		".section .fixup,\"ax\"\n"
		"3:\n\t"
		"mov.l	4f,%1\n\t"
		"jmp	@%1\n\t"
		" mov	%5,%0\n"
		".balign 4\n"
		"4:	.long 2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.balign 4\n"
		"	.long 1b,3b\n"
		".previous"
		: "=&r" (res), "=&z" (__dummy)
		: "0" (__s), "r" (__s), "r" (__s+__n), "i" (-EFAULT));
	return res;
}

#endif CONFIG_ZAURUS

extern __inline__ long strnlen_user(const char *s, long n)
{
	if(!access_ok(VERIFY_READ, s, n))
		return 0;
	else
		return __strnlen_user(s, n);
}

struct exception_table_entry
{
	unsigned long insn, fixup;
};

/* Returns 0 if exception not found and fixup.unit otherwise.  */
extern unsigned long search_exception_table(unsigned long addr);

/* Returns the new pc */
#define fixup_exception(map_reg, fixup_unit, pc)                \
({                                                              \
	fixup_unit;                                             \
})

#endif /* __ASM_SH_UACCESS_H */
